import pytest
from patent_client.uspto_ptab import PtabDocument
from patent_client.uspto_ptab import PtabTrial


class TestPtab:
    def test_can_get_case(self):
        case = PtabTrial.objects.get("IPR2016-00831")
        assert case.patent_owner_name == "Silicon Genesis Corporation"

    def test_can_get_set_of_cases(self):
        cases = PtabTrial.objects.filter(
            patent_owner_name="Silicon Genesis Corporation"
        ).order_by("trial_number")
        assert cases.count() == 3
        assert list(cases.all())[0].trial_number == "IPR2016-00831"
        assert len(list(cases)) == 3

    def test_can_get_case_documents(self):
        case = PtabTrial.objects.get("IPR2016-00832")
        assert case.documents.count() == 77
        assert (
            case.documents.first().title
            == "U.S. Provisional Application No. 60/046,276"
        )
        assert case.documents.values_list("title", flat=True)[:5] == [
            "U.S. Provisional Application No. 60/046,276",
            "U.S. Patent No. 5,374,564",
            "L. Wegmann, Historical Perspective and Future Trends for Ion Implantation Systems (1981)",
            "U.S. Patent No. 5,559,043",
            "Prosecution History of U.S. Patent No. 6,013,563",
        ]

    def test_can_get_case_from_document(self):
        doc = PtabDocument.objects.get(230910)
        assert doc.trial.trial_number == "IPR2016-00831"

    def test_can_download_document(self, tmpdir):
        doc = PtabDocument.objects.get(230910)
        doc.download(path=tmpdir)
        expected_doc = tmpdir / "U.S. Provisional Application No. 60_046,276.pdf"
        assert expected_doc.exists()



import json
import mimetypes
import os
import shutil

import inflection
import requests
from patent_client import CACHE_BASE
from patent_client.util import Manager
from patent_client.util import Model
from patent_client.util import hash_dict
from patent_client.util import one_to_many
from patent_client.util import one_to_one

CACHE_DIR = CACHE_BASE / "ptab"
CACHE_DIR.mkdir(exist_ok=True)

session = requests.Session()


class PtabManager(Manager):
    page_size = 25

    def get_item(self, key):
        offset = int(key / self.page_size) * self.page_size
        position = key % self.page_size
        data = self.request(dict(offset=offset))
        results = data["results"]
        return self.get_obj_class()(results[position])

    def __len__(self):
        response = self.request()
        response_data = response
        count = response_data["metadata"]["count"]
        return count

    def request(self, params=dict()):
        params = {
            inflection.camelize(k, uppercase_first_letter=False): v
            for (k, v) in {**self.filter_params, **dict(sort=self.sort_params)}.items()
        }
        fname = CACHE_DIR / f"{self.__class__}-{hash_dict(params)}.json"
        if not fname.exists():
            response = session.get(self.base_url, params=params)
            with open(fname, "w") as f:
                json.dump(response.json(), f, indent=True)
        return json.load(open(fname))


class PtabTrialManager(PtabManager):
    base_url = "https://ptabdata.uspto.gov/ptab-api/trials"
    obj_class = "patent_client.uspto_ptab.PtabTrial"
    primary_key = "trial_number"


class PtabDocumentManager(PtabManager):
    base_url = "https://ptabdata.uspto.gov/ptab-api/documents"
    obj_class = "patent_client.uspto_ptab.PtabDocument"
    primary_key = "id"


class PtabTrial(Model):
    objects = PtabTrialManager()
    us_application = one_to_one(
        "patent_client.USApplication", appl_id="application_number"
    )
    documents = one_to_many("patent_client.PtabDocument", trial_number="trial_number")

    def __repr__(self):
        return f"<PtabTrial(trial_number={self.trial_number})>"


class PtabDocument(Model):
    objects = PtabDocumentManager()
    trial = one_to_one("patent_client.PtabTrial", trial_number="trial_number")

    def download(self, path="."):
        url = self.links[1]["href"]
        extension = mimetypes.guess_extension(self.media_type)
        base_name = self.title.replace("/", "_") + extension
        cdir = os.path.join(CACHE_DIR, self.trial_number)
        os.makedirs(cdir, exist_ok=True)
        cname = os.path.join(cdir, base_name)
        oname = os.path.join(path, base_name)
        if not os.path.exists(cname):
            response = session.get(url, stream=True)
            with open(cname, "wb") as f:
                for chunk in response.iter_content(1024):
                    f.write(chunk)
        shutil.copy(cname, oname)

    def __repr__(self):
        return f"<PtabDocument(title={self.title})>"



from collections import OrderedDict

import pytest
from patent_client import Inpadoc
from patent_client import PtabTrial
from patent_client import USApplication


class TestIntegration:
    def test_can_get_trials(self):
        pat_no = "8118221"
        app = USApplication.objects.get(patent_number=pat_no)
        assert list(app.trials.values("trial_number")) == [
            OrderedDict([("trial_number", "CBM2014-00102")]),
            OrderedDict([("trial_number", "CBM2014-00103")]),
            OrderedDict([("trial_number", "CBM2014-00194")]),
            OrderedDict([("trial_number", "CBM2014-00199")]),
            OrderedDict([("trial_number", "CBM2015-00015")]),
            OrderedDict([("trial_number", "CBM2015-00117")]),
            OrderedDict([("trial_number", "CBM2015-00126")]),
            OrderedDict([("trial_number", "CBM2015-00130")]),
        ]

    def test_can_get_app_from_trials(self):
        trial = PtabTrial.objects.get("CBM2014-00102")
        print(trial.us_application.patent_title)
        assert trial.us_application.patent_title == "DATA STORAGE AND ACCESS SYSTEMS"

    def test_can_get_inpadoc(self):
        pat_no = "8118221"
        app = USApplication.objects.get(patent_number=pat_no)
        assert app.inpadoc[0].title == "Data Storage and Access Systems"

    def test_can_get_us_app_from_inpadoc(self):
        app = Inpadoc.objects.get("US9231827B2")
        assert (
            app.us_application.patent_title
            == "FORMALIZING, DIFFUSING AND ENFORCING POLICY ADVISORIES AND MONITORING POLICY COMPLIANCE IN THE MANAGEMENT OF NETWORKS"
        )

    def test_can_get_assignments_from_application(self):
        app = USApplication.objects.get("13842218")
        assert list(app.assignments.values_list("id", flat=True)) == [
            "36385-377",
            "36408-10",
            "37418-226",
        ]



import os

from patent_client.epo_ops import CACHE_DIR
from patent_client.util import Manager
from patent_client.util import Model
from patent_client.util import one_to_one
from PyPDF2 import PdfFileMerger

from .ops import EpoConnector
from .ops import InpadocConnector

inpadoc_connector = InpadocConnector()
epo_connector = EpoConnector()


class InpadocManager(Manager):
    obj_class = "patent_client.epo_ops.models.Inpadoc"
    primary_key = "publication"
    connector = inpadoc_connector

    def __init__(self, *args, **kwargs):
        super(InpadocManager, self).__init__(*args, **kwargs)
        self.pages = dict()

    def __len__(self):
        """Total number of results"""
        if "application" in self.filter_params or "publication" in self.filter_params:
            return len(self.get_by_number())
        else:
            return self.connector.get_search_length(self.filter_params)

    def get_by_number(self):
        if "publication" in self.kwargs:
            number = self.kwargs["publication"]
            if not isinstance(number, list):
                number = number[0]
            doc_db = self.connector.original_to_docdb(number, "publication")
        elif "application" in self.kwargs:
            number = self.kwargs["application"]
            if isinstance(number, list):
                number = number[0]
            doc_db = self.connector.original_to_docdb(number, "application")
        docs = self.connector.bib_data(doc_db)
        return docs

    def get_item(self, key):
        if "publication" in self.kwargs or "application" in self.kwargs:
            docs = self.get_by_number()
            return Inpadoc(docs[key])
        else:
            # Search Iterator
            doc_db = self.connector.get_search_item(key, self.filter_params)
            return Inpadoc(self.connector.bib_data(doc_db)[0])


class InpadocFullTextManager(InpadocManager):
    def get(self, doc_db):
        data = {
            "description": self.connector.description(doc_db),
            "claims": self.connector.claims(doc_db),
            "doc_db": doc_db,
        }
        return InpadocFullText(data)


class Inpadoc(Model):
    objects = InpadocManager()
    full_text = one_to_one(
        "patent_client.epo_ops.models.InpadocFullText", doc_db="doc_db"
    )
    us_application = one_to_one(
        "patent_client.USApplication", appl_id="original_application_number"
    )

    def __repr__(self):
        return f"<Inpadoc(publication={self.publication})>"

    @property
    def legal(self):
        if not hasattr(self, "_legal"):
            data = inpadoc_connector.legal(self.doc_db)
            self._legal = data
        return self._legal

    @property
    def images(self):
        if not hasattr(self, "_images"):
            data = inpadoc_connector.images(self.doc_db)
            data["doc_db"] = self.doc_db
            self._images = InpadocImages(data)
        return self._images

    @property
    def family(self):
        for doc_db in inpadoc_connector.family(self.doc_db):
            yield Inpadoc(inpadoc_connector.bib_data(doc_db)[0])


class InpadocFullText(Model):
    objects = InpadocFullTextManager()


class InpadocImages(Model):
    objects = InpadocManager()

    def download(self, path="."):
        """Download each page of images, and then consolidate into a single PDF
        Args:
            path: str(base path for file)
        """
        dirname = (
            CACHE_DIR
            / f'{self.doc_db.doc_type}-{self.doc_db.country}{self.doc_db.number}{self.doc_db.kind if self.doc_db.kind else ""}'
        )
        dirname.mkdir(exist_ok=True)
        for i in range(1, self.num_pages + 1):
            fname = dirname / ("page-" + str(i).rjust(6, "0") + ".pdf")
            if not fname.exists():
                inpadoc_connector.pdf_request(fname, self.url, params={"Range": i})

        pages = list(sorted(os.listdir(dirname)))
        out_file = PdfFileMerger()
        for p in pages:
            out_file.append(os.path.join(dirname, p))
        out_fname = os.path.join(
            path, self.doc_db.country + self.doc_db.number + ".pdf"
        )
        out_file.write(out_fname)


class EpoManager(Manager):
    """
    EPO Manager
    This is a manager class to the EPO register. Retrieves information about
    EPO patents and applications
    """

    obj_class = "patent_client.epo_ops.models.Epo"

    def get(self, number=None, doc_type="publication"):
        epodoc = epo_connector.original_to_epodoc(number, doc_type)
        bib_data = epo_connector.bib_data(epodoc)
        return Epo(bib_data)


class Epo(Model):
    objects = EpoManager()

    def __repr__(self):
        return (
            f"<Epo(number={self.number}, kind={self.kind}, doc_type={self.doc_type})>"
        )

    @property
    def procedural_steps(self):
        return epo_connector.procedural_steps(self.epodoc)



import datetime

from patent_client import ITCInvestigation


class TestItcEdis:
    def test_can_get_investigation(self):
        inv = ITCInvestigation.objects.get("337-TA-1025")
        assert inv.data == {
            "phase": "Violation",
            "number": "337-1025",
            "status": "Active",
            "title": "Certain Silicon-on-Insulator Wafers; Inv. No. 337-TA-1025",
            "type": "Sec 337",
            "doc_list_url": "https://edis.usitc.gov/data/document?investigationNumber=337-1025&investigationPhase=Violation",
            "docket_number": "3153",
        }

    def test_can_get_documents(self):
        inv = ITCInvestigation.objects.get("337-TA-966")
        print(inv.documents[0].data)
        assert inv.documents[0].data == {
            "investigation_number": "337-966",
            "type": "Voting Sheet",
            "title": None,
            "security": "Public",
            "filing_org": "USITC",
            "filed_by": "Lisa R. Barton",
            "filed_on_behalf_of": "Office of the Secretary",
            "action_jacket_control_number": "GC-16-249",
            "memorandum_control_number": None,
            "attachment_url": "https://edis.usitc.gov/data/attachment/600197",
            "date": datetime.date(2016, 12, 2),
            "last_modified": "2017/05/23 15:43:34",
            "id": "600197",
        }
        assert inv.documents[0].investigation.number == "337-966"

    def test_can_get_attachments(self, tmpdir):
        inv = ITCInvestigation.objects.get("337-TA-1025")
        assert inv.documents[15].attachments[0].data == {
            "id": "1163223",
            "document_id": "605365",
            "title": "605365",
            "size": "70956",
            "file_name": "/var/edis/data/sibyl/605365/605365.pdf",
            "pages": "3",
            "created_date": datetime.date(2017, 3, 13),
            "last_modified_date": datetime.date(2017, 3, 13),
            "download_url": "https://edis.usitc.gov/data/download/605365/1163223",
        }
        inv.documents[15].attachments[0].download(tmpdir)
        expected_doc = (
            tmpdir
            / "Granting Respondent's Unopposed Motion for Leave to Designate a Substitute Expert - 605365.pdf"
        )
        exists = expected_doc.exists()
        assert exists



import json
import math
import os
import re
from collections import namedtuple
from hashlib import md5

import requests
from lxml import etree as ET
from patent_client import SETTINGS
from patent_client.epo_ops import CACHE_DIR

session = requests.Session()

CLIENT_SETTINGS = SETTINGS["EpoOpenPatentServices"]
if os.environ.get("EPO_KEY", False):
    KEY = os.environ["EPO_KEY"]
    SECRET = os.environ["EPO_SECRET"]
else:
    KEY = CLIENT_SETTINGS["ApiKey"]
    SECRET = CLIENT_SETTINGS["Secret"]

# XML Namespaces for ElementTree
NS = {
    "ops": "http://ops.epo.org",
    "epo": "http://www.epo.org/exchange",
    "ft": "http://www.epo.org/fulltext",
    "reg": "http://www.epo.org/register",
}

SEARCH_FIELDS = {
    "title": "title",
    "abstract": "abstract",
    "title_and_abstract": "titleandabstract",
    "inventor": "inventor",
    "applicant": "applicant",
    "inventor_or_applicant": "inventorandapplicant",
    "publication": "publicationnumber",
    "epodoc_publication": "spn",
    "application": "applicationnumber",
    "epodoc_application": "sap",
    "priority": "prioritynumber",
    "epodoc_priority": "spr",
    "number": "num",  # Pub, App, or Priority Number
    "publication_date": "publicationdate",  # yyyy, yyyyMM, yyyyMMdd, yyyy-MM, yyyy-MM-dd
    "citation": "citation",
    "cited_in_examination": "ex",
    "cited_in_opposition": "op",
    "cited_by_applicant": "rf",
    "other_citation": "oc",
    "family": "famn",
    "cpc_class": "cpc",
    "ipc_class": "ipc",
    "ipc_core_invention_class": "ci",
    "ipc_core_additional_class": "cn",
    "ipc_advanced_class": "ai",
    "ipc_advanced_additional_class": "an",
    "ipc_core_class": "c",
    "classification": "cl",  # IPC or CPC Class
    "full_text": "txt",  # title, abstract, inventor and applicant
}

EpoDoc = namedtuple("EpoDoc", ["number", "kind", "date", "doc_type"])
DocDB = namedtuple("DocDB", ["country", "number", "kind", "date", "doc_type"])
whitespace_re = re.compile(" +")
country_re = re.compile(r"^[A-Z]{2}")
ep_case_re = re.compile(r"EP(?P<number>[\d]+)(?P<kind>[A-Z]\d)?")


class OPSException(Exception):
    pass


class OPSAuthenticationException(OPSException):
    pass


class OpenPatentServicesConnector:
    def authenticate(self, key=None, secret=None):
        auth_url = "https://ops.epo.org/3.2/auth/accesstoken"
        global KEY, SECRET
        if key:
            KEY = key
            SECRET = secret

        response = session.post(
            auth_url, auth=(KEY, SECRET), data={"grant_type": "client_credentials"}
        )

        if not response.ok:
            raise OPSAuthenticationException()

        access_token = response.json()["access_token"]
        session.headers["Authorization"] = "Bearer " + access_token

    def pdf_request(self, fname, url, params=dict()):
        if os.path.exists(fname):
            return
        else:
            with open(fname, "wb") as f:
                response = self.request(url, params, stream=True)
                f.write(response.raw.read())

    def request(self, url, params=dict(), stream=False):
        retry = 0
        while retry < 3:
            response = session.get(url, params=params, stream=stream)
            if response.ok:
                return response
            elif response.status_code in (400, 403):
                self.authenticate()
            elif not response.ok:
                tree = ET.fromstring(response.text.encode("utf-8"))
                code = tree.find("./ops:code", NS).text
                message = tree.find("./ops:message", NS).text
                details = tree.find("./ops:details", NS)
                if details is not None:
                    details = "".join(details.itertext())
                else:
                    details = "<None>"
                raise OPSException(
                    f"{response.status_code} - {code} - {message}\nDETAILS: {details} \nURL: {response.request.url}"
                )
            retry += 1
        raise OPSException("Max Retries Exceeded!")

    def xml_request(self, url, params=dict()):
        param_hash = md5(json.dumps(params, sort_keys=True).encode("utf-8")).hexdigest()
        fname = os.path.join(
            CACHE_DIR, f"{url[37:].replace('/', '_')}{param_hash if params else ''}.xml"
        )
        if os.path.exists(fname):
            return open(fname).read()
        response = self.request(url, params)
        text = response.text
        with open(fname, "w") as f:
            f.write(text)
        return text

    def original_to_docdb(self, number, doc_type):
        if "PCT" in number:
            return self.pct_to_docdb(number)
        country = country_re.search(number)
        if country:
            country = country.group(0)
            number = number[2:]
        else:
            country = "US"
            number = number

        url = f"http://ops.epo.org/3.2/rest-services/number-service/{doc_type}/original/{country}.({number})/docdb"
        text = self.xml_request(url)
        tree = ET.fromstring(text.encode("utf-8"))
        output = tree.find("./ops:standardization/ops:output", NS)

        if doc_type == "application":
            app_ref = output.find("./ops:application-reference/epo:document-id", NS)
            return self.docdb_number(app_ref, doc_type)
        elif doc_type == "publication":
            pub_ref = output.find("./ops:publication-reference/epo:document-id", NS)
            return self.docdb_number(pub_ref, doc_type)

    def original_to_epodoc(self, number, doc_type):
        url = f"http://ops.epo.org/3.2/rest-services/number-service/{doc_type}/original/{number})/epodoc"
        text = self.xml_request(url)
        tree = ET.fromstring(text.encode("utf-8"))
        output = tree.find("./ops:standardization/ops:output", NS)

        if doc_type == "application":
            app_ref = output.find("./ops:application-reference/epo:document-id", NS)
            return self.epodoc_number(app_ref, doc_type)
        elif doc_type == "publication":
            pub_ref = output.find("./ops:publication-reference/epo:document-id", NS)
            return self.epodoc_number(pub_ref, doc_type)

    def pct_to_docdb(self, number):
        _, country_year, number = number.split("/")
        country, year = country_year[:2], country_year[2:]
        if len(year) == 2:
            if int(year) > 50:
                year = "19" + year
            else:
                year = "20" + year

        # DocDB format changed in 2004:
        # Pre-2003 - CCyynnnnnW
        # Post-2003 - CCccyynnnnnnW
        if int(year) >= 2004:
            case_number = year + number.rjust(6, "0")
        else:
            case_number = year[2:] + number.rjust(5, "0")
        return DocDB(country, case_number, "W", None, "application")

    def docdb_number(self, el, doc_type):
        raw_data = {
            "country": el.find("./epo:country", NS),
            "number": el.find("./epo:doc-number", NS),
            "kind": el.find("./epo:kind", NS),
            "date": el.find("./epo:date", NS),
        }
        for k, v in raw_data.items():
            if v is not None:
                raw_data[k] = v.text

        raw_data["doc_type"] = doc_type

        return DocDB(**raw_data)

    def epodoc_number(self, el, doc_type):
        raw_data = {
            "number": el.find("./epo:doc-number", NS),
            "kind": el.find("./epo:kind", NS),
            "date": el.find("./epo:date", NS),
        }
        for k, v in raw_data.items():
            if v is not None:
                raw_data[k] = v.text

        raw_data["doc_type"] = doc_type

        return EpoDoc(**raw_data)

    def cpc_class(self, el):
        keys = "section, class, subclass, main-group, subgroup".split(", ")
        data = {k: el.find("./epo:" + k, NS) for k in keys}
        data = {k: v.text for (k, v) in data.items() if v is not None}
        return f"{data.get('section', '')}{data.get('class', '')}{data.get('subclass', '')} {data.get('main-group', '')}/{data.get('subgroup', '')}"


class InpadocConnector(OpenPatentServicesConnector):
    page_size = 25
    search_url = "http://ops.epo.org/3.2/rest-services/published-data/search"

    def xml_data(self, pub, data_kind):
        case_number = f'{pub.country}.{pub.number}{"." + pub.kind if pub.kind else ""}'
        if data_kind not in (
            "biblio",
            "description",
            "claims",
            "images",
            "legal",
            "family",
        ):
            raise OPSException(data_kind + " is not a valid data kind")
        if data_kind == "legal":
            legal_url = "http://ops.epo.org/3.2/rest-services/legal/{doc_type}/docdb/{case_number}/"
            url = legal_url.format(case_number=case_number, doc_type=pub.doc_type)
        elif data_kind == "family":
            family_url = "http://ops.epo.org/3.2/rest-services/family/{doc_type}/docdb/{case_number}/"
            url = family_url.format(case_number=case_number, doc_type=pub.doc_type)
        else:
            data_url = "http://ops.epo.org/3.2/rest-services/published-data/{doc_type}/docdb/{case_number}/{data_kind}"
            url = data_url.format(
                case_number=case_number, doc_type=pub.doc_type, data_kind=data_kind
            )
        text = self.xml_request(url)
        tree = ET.fromstring(text.encode("utf-8"))
        return tree

    def get_search_page(self, page_number, query):
        start = (page_number - 1) * self.page_size + 1
        end = start + self.page_size - 1
        params = {**query, **{"Range": f"{start}-{end}"}}
        text = self.xml_request(self.search_url, params)
        tree = ET.fromstring(text.encode("utf-8"))
        results = tree.find(".//ops:search-result", NS)
        return [
            self.docdb_number(el.find("./epo:document-id", NS), "publication")
            for el in results
        ]

    def get_search_item(self, key, query_dict=None):
        query = self.create_query(query_dict)
        page_num = math.floor(key / self.page_size) + 1
        line_num = key % self.page_size
        page = self.get_search_page(page_num, query)
        return page[line_num]

    def get_search_length(self, query_dict=None):
        query = self.create_query(query_dict)
        params = {**query, **{"Range": f"1-{self.page_size}"}}
        text = self.xml_request(self.search_url, params)
        tree = ET.fromstring(text.encode("utf-8"))
        return int(tree.find(".//ops:biblio-search", NS).attrib["total-result-count"])

    def create_query(self, query_dict):
        if "cql_query" in query_dict:
            return dict(q=query_dict["cql_query"])
        query = ""
        for keyword, value in query_dict.items():
            if len(value.split()) > 1:
                value = f'"{value}"'
            query += SEARCH_FIELDS[keyword] + "=" + value
        return dict(q=query)

    def parse_citation(self, el):
        phase = el.attrib["cited-phase"]
        cited_by = el.attrib["cited-by"]
        office = el.attrib.get("office", "")
        pat_cite = el.find("./epo:patcit", NS)
        if pat_cite is not None:
            citation = dict(
                self.docdb_number(
                    pat_cite.find('./epo:document-id[@document-id-type="docdb"]', NS),
                    "publication",
                )._asdict()
            )
        else:
            citation = el.find("./epo:nplcit/epo:text", NS).text
        category = (
            el.find("./epo:category", NS).text
            if el.find("./epo:category", NS) is not None
            else ""
        )
        relevant_claims = el.find("./epo:rel-claims", NS)
        relevant_passages = [
            e.text for e in el.findall("./epo:rel-passage/epo:passage", NS)
        ]
        return {
            "phase": phase,
            "cited_by": cited_by,
            "office": office,
            "citation": citation,
            "category": category,
            "relevant_claims": relevant_claims.text
            if relevant_claims is not None
            else "",
            "relevant_passages": relevant_passages,
        }

    def legal(self, doc_db):
        tree = self.xml_data(doc_db, "legal")
        legal_events = tree.findall(".//ops:legal", NS)
        output = list()
        for le in legal_events:
            row = dict()
            row["description"] = le.attrib["desc"]
            row["explanation"] = le.attrib["infl"]
            row["code"] = le.attrib["code"].strip()
            row["date"] = le.find("./ops:L007EP", NS).text
            output.append(row)
        output = list(sorted(output, key=lambda x: x["date"]))
        return output

    def bib_data(self, doc_db):
        # NOTE: For EP cases with search reports, there is citation data in bib data
        # We just currently are not parsing it
        tree = self.xml_data(doc_db, "biblio")
        documents = tree.findall("./epo:exchange-documents/epo:exchange-document", NS)

        data_items = list()
        for document in documents:
            data = dict()
            data["family_id"] = document.attrib["family-id"]

            bib_data = document.find("./epo:bibliographic-data", NS)

            title = bib_data.find("./epo:invention-title[@lang='en']", NS)
            if title is None:
                title = bib_data.find("./epo:invention-title", NS)
            if title is not None:
                data["title"] = title.text.strip()
            else:
                data["title"] = ""

            pub_data = bib_data.find(
                './epo:publication-reference/epo:document-id[@document-id-type="docdb"]',
                NS,
            )
            pub_data = self.docdb_number(pub_data, "publication")
            data["doc_db"] = pub_data
            data["publication"] = pub_data.country + pub_data.number + pub_data.kind
            data["publication_date"] = pub_data.date
            data["country"] = pub_data.country

            app_data = bib_data.find(
                './epo:application-reference/epo:document-id[@document-id-type="docdb"]',
                NS,
            )
            app_data = dict(self.docdb_number(app_data, "application")._asdict())
            data["application"] = app_data["country"] + app_data["number"]
            data["filing_date"] = app_data.get("date", None)

            original_app_data = bib_data.find(
                './epo:application-reference/epo:document-id[@document-id-type="original"]',
                NS,
            )
            if original_app_data is not None:
                data["original_application_number"] = original_app_data.find(
                    "./epo:doc-number", NS
                ).text

            data["intl_class"] = [
                whitespace_re.sub("", e.text)
                for e in bib_data.findall(
                    "./epo:classifications-ipcr/epo:classification-ipcr/epo:text", NS
                )
            ]

            data["cpc_class"] = [
                self.cpc_class(el)
                for el in bib_data.findall(
                    "./epo:patent-classifications/epo:patent-classification", NS
                )
            ]

            data["priority_claims"] = [
                e.text
                for e in bib_data.findall(
                    './epo:priority-claims/epo:priority-claim/epo:document-id[@document-id-type="original"]/epo:doc-number',
                    NS,
                )
            ]

            parties = bib_data.find("./epo:parties", NS)
            data["applicants"] = [
                e.text
                for e in parties.findall(
                    './epo:applicants/epo:applicant[@data-format="original"]/epo:applicant-name/epo:name',
                    NS,
                )
            ]
            data["inventors"] = [
                e.text
                for e in parties.findall(
                    './epo:inventors/epo:inventor[@data-format="original"]/epo:inventor-name/epo:name',
                    NS,
                )
            ]
            abstract = document.find('./epo:abstract[@lang="en"]', NS)
            data["abstract"] = (
                "".join(t for t in abstract.itertext()).strip()
                if abstract is not None
                else ""
            )

            refs_cited = bib_data.findall("./epo:references-cited/epo:citation", NS)
            data["references_cited"] = [self.parse_citation(c) for c in refs_cited]
            data_items.append(data)
        return data_items

    def description(self, doc_db):
        tree = self.xml_data(doc_db, "description")
        description = tree.find(
            "./ft:fulltext-documents/ft:fulltext-document/ft:description", NS
        )
        text = "\n".join(description.itertext()).strip()
        return text

    def claims(self, doc_db):
        tree = self.xml_data(doc_db, "claims")
        return [
            "".join(e.itertext()).strip()
            for e in tree.findall(
                './ft:fulltext-documents/ft:fulltext-document/ft:claims[@lang="EN"]/ft:claim/ft:claim-text',
                NS,
            )
        ]

    def images(self, doc_db):
        if doc_db.doc_type == "application":
            return dict()
        tree = self.xml_data(doc_db, "images")
        images = tree.find(
            './ops:document-inquiry/ops:inquiry-result/ops:document-instance[@desc="FullDocument"]',
            NS,
        )
        return {
            "url": "http://ops.epo.org/rest-services/" + images.attrib["link"] + ".pdf",
            "num_pages": int(images.attrib["number-of-pages"]),
            "sections": {
                e.attrib["name"]: int(e.attrib["start-page"])
                for e in images.findall("./ops:document-section", NS)
            },
        }

    def family(self, doc_db):
        tree = self.xml_data(doc_db, "family")
        doc_db_list = list()
        for el in tree.findall("./ops:patent-family/ops:family-member", NS):
            pub_ref = el.find(
                './epo:publication-reference/epo:document-id[@document-id-type="docdb"]',
                NS,
            )
            if pub_ref is not None:
                doc_db_list.append(self.docdb_number(pub_ref, "publication"))
            else:
                app_ref = el.find(
                    './epo:application-reference/epo:document-id[@document-id-type="docdb"]',
                    NS,
                )
                doc_db_list.append(self.docdb_number(app_ref, "application"))
        return doc_db_list


class EpoConnector(OpenPatentServicesConnector):
    def xml_data(self, pub, data_kind):
        """
        Acceptable kinds are "biblio", "events", "procedural-steps"
        """
        case_number = f"EP.{pub.number[2:]}"
        if pub.kind:
            case_number += "." + pub.kind

        register_url = "http://ops.epo.org/3.2/rest-services/register/{doc_type}/epodoc/{case_number}/"
        if data_kind not in ("biblio", "events", "procedural-steps"):
            raise ValueError(data_kind + " is not a valid data kind")
        url = (
            register_url.format(case_number=case_number, doc_type=pub.doc_type)
            + data_kind
        )
        text = self.xml_request(url)
        tree = ET.fromstring(text.encode("utf-8"))
        return tree

    def status(self, el):
        return {
            "description": el.text,
            "code": el.attrib["status-code"],
            "date": el.attrib["change-date"],
        }

    def priority_claim(self, el):
        return {
            "kind": el.attrib["kind"],
            "number": el.find("./reg:doc-number", NS).text,
            "date": el.find("./reg:date", NS).text,
        }

    def docid(self, el):
        doc_id = el.find(".//reg:document-id", NS)
        raw = {
            "country": doc_id.find("./reg:country", NS),
            "number": doc_id.find("./reg:doc-number", NS),
            "kind": doc_id.find("./reg:kind", NS),
            "date": doc_id.find("./reg:date", NS),
        }
        return {k: v.text for (k, v) in raw.items() if v is not None}

    def party(self, el):
        addr_book = el.find("./reg:addressbook", NS)
        name = addr_book.find("./reg:name", NS)
        address = addr_book.find("./reg:address", NS)
        return {
            "name": name.text,
            "address": "\n".join(
                [t.strip() for t in address.itertext() if t.strip()]
            ).strip(),
        }

    def citation(self, el):
        phase = el.attrib["cited-phase"]
        office = el.attrib.get("office", "")
        pat_cite = el.find("./reg:patcit", NS)
        if pat_cite is not None:
            citation = self.docid(pat_cite)
        else:
            citation = el.find("./reg:nplcit/reg:text", NS).text
        category = (
            el.find("./reg:category", NS).text
            if el.find("./reg:category", NS) is not None
            else ""
        )
        relevant_passages = pat_cite.find("./reg:text", NS).text
        return {
            "phase": phase,
            "office": office,
            "citation": citation,
            "category": category,
            "relevant_passages": relevant_passages,
        }

    def step(self, el):
        date = el.find("./reg:procedural-step-date/reg:date", NS)
        return {
            "phase": el.attrib["procedure-step-phase"],
            "description": " - ".join(
                [e.text for e in el.findall("./reg:procedural-step-text", NS)]
            ),
            "date": date.text.strip() if date is not None else None,
            "code": el.find("./reg:procedural-step-code", NS).text.strip(),
        }

    def procedural_steps(self, epodoc):
        tree = self.xml_data(epodoc, "procedural-steps")
        doc = tree.find(".//reg:register-document", NS)
        return [self.step(el) for el in doc.find("./reg:procedural-data", NS)]

    def bib_data(self, epodoc):
        tree = self.xml_data(epodoc, "biblio")
        doc = tree.find(".//reg:register-document", NS)
        bib = doc.find("./reg:bibliographic-data", NS)
        parties = bib.find("./reg:parties", NS)
        output = dict()
        output["number"] = epodoc.number
        output["kind"] = epodoc.kind
        output["case_number"] = epodoc.number + epodoc.kind
        output["epodoc"] = epodoc
        output["doc_type"] = epodoc.doc_type
        output["status"] = [
            self.status(el) for el in doc.find("./reg:ep-patent-statuses", NS)
        ]
        output["publications"] = [
            self.docid(el) for el in bib.findall("./reg:publication-reference", NS)
        ]
        output["intl_class"] = [
            "".join(el.itertext()).strip()
            for el in bib.find("./reg:classifications-ipcr", NS)
        ]
        output["applications"] = [
            self.docid(el) for el in bib.findall("./reg:application-reference", NS)
        ]
        output["filing_language"] = bib.find("./reg:language-of-filing", NS).text
        output["priority_claims"] = [
            self.priority_claim(el) for el in bib.find("reg:priority-claims", NS)
        ]
        output["applicants"] = [
            self.party(el) for el in parties.find("./reg:applicants", NS)
        ]
        output["inventors"] = [
            self.party(el) for el in parties.find("./reg:inventors", NS)
        ]
        output["designated_states"] = [
            c.strip()
            for c in bib.find("./reg:designation-of-states", NS).itertext()
            if c.strip()
        ]
        output["title"] = bib.find('./reg:invention-title[@lang="en"]', NS).text
        output["citations"] = [
            self.citation(el) for el in bib.find("./reg:references-cited", NS)
        ]
        return output



import datetime
from collections import OrderedDict

from patent_client.uspto_exam_data import USApplication


class TestPatentExaminationData:
    """
    def test_search_by_customer_number(self):
        result = USApplication.objects.search(app_cust_number="70155")
        print(json.dumps(result))
        assert False
    """

    def test_get_by_pub_number(self):
        pub_no = "US20060127129A1"
        app = USApplication.objects.get(app_early_pub_number=pub_no)
        assert app.patent_title == "ELECTROPHOTOGRAPHIC IMAGE FORMING APPARATUS"

    def test_get_by_pat_number(self):
        pat_no = 6095661
        app = USApplication.objects.get(patent_number=pat_no)
        assert app.patent_title == "METHOD AND APPARATUS FOR AN L.E.D. FLASHLIGHT"

    def test_get_by_application_number(self):
        app_no = "15145443"
        app = USApplication.objects.get(app_no)
        assert (
            app.patent_title
            == "Suction and Discharge Lines for a Dual Hydraulic Fracturing Unit"
        )

    def test_get_many_by_application_number(self):
        app_nos = ["14971450", "15332765", "13441334", "15332709", "14542000"]
        data = USApplication.objects.get_many(*app_nos)
        data_list = list(data)
        assert len(data_list) == 5
        assert (
            data_list[4].patent_title
            == "MOBILE, MODULAR, ELECTRICALLY POWERED SYSTEM FOR USE IN FRACTURING UNDERGROUND FORMATIONS"
        )

        assert list(data.values_list("patent_title", flat=True)) == [
            "SYSTEM AND METHOD FOR DEDICATED ELECTRIC SOURCE FOR USE IN FRACTURING UNDERGROUND FORMATIONS USING LIQUID PETROLEUM GAS",
            "ELECTRIC BLENDER SYSTEM, APPARATUS AND METHOD FOR USE IN FRACTURING UNDERGROUND FORMATIONS USING LIQUID PETROLEUM GAS",
            "MOBILE ELECTRIC POWER GENERATION FOR HYDRAULIC FRACTURING OF SUBSURFACE GEOLOGICAL FORMATIONS",
            "MOBILE, MODULAR, ELECTRICALLY POWERED SYSTEM  FOR USE IN FRACTURING UNDERGROUND FORMATIONS",
            "MOBILE, MODULAR, ELECTRICALLY POWERED SYSTEM FOR USE IN FRACTURING UNDERGROUND FORMATIONS",
        ]
        assert list(data.values("appl_id", "app_filing_date")) == [
            OrderedDict(
                [
                    ("appl_id", "15332709"),
                    ("app_filing_date", datetime.date(2016, 10, 24)),
                ]
            ),
            OrderedDict(
                [
                    ("appl_id", "15332765"),
                    ("app_filing_date", datetime.date(2016, 10, 24)),
                ]
            ),
            OrderedDict(
                [
                    ("appl_id", "14971450"),
                    ("app_filing_date", datetime.date(2015, 12, 16)),
                ]
            ),
            OrderedDict(
                [
                    ("appl_id", "13441334"),
                    ("app_filing_date", datetime.date(2012, 4, 6)),
                ]
            ),
            OrderedDict(
                [
                    ("appl_id", "14542000"),
                    ("app_filing_date", datetime.date(2014, 11, 14)),
                ]
            ),
        ]

    def test_values_many_by_application(self):
        app_nos = ["14971450", "15332765", "13441334", "15332709", "14542000"]
        data = USApplication.objects.filter(*app_nos)
        print(list(data.values("inventors__0__nameLineOne")))
        assert list(data.values("inventors__0__nameLineOne")) == [
            OrderedDict([("inventors_0_nameLineOne", "Todd  Coli")]),
            OrderedDict([("inventors_0_nameLineOne", "Schelske, Eldon")]),
            OrderedDict([("inventors_0_nameLineOne", "Jeffrey G. Morris")]),
            OrderedDict([("inventors_0_nameLineOne", "Todd  Coli")]),
            OrderedDict([("inventors_0_nameLineOne", "Todd  Coli")]),
        ]

    def test_search_patex_by_assignee(self):
        data = USApplication.objects.filter(first_named_applicant="LogicBlox")
        assert data.order_by("appl_id").values_list("patent_title", flat=True)[:5] == [
            "MAINTENANCE OF ACTIVE DATABASE QUERIES",
            "LEAPFROG TREE-JOIN",
            "SALIENT SAMPLING FOR QUERY SIZE ESTIMATION",
            "TRANSACTION REPAIR",
            "LEAPFROG TREE-JOIN",
        ]

    def test_get_many_by_publication_number(self):
        nos = [
            "US20080034424A1",
            "US20100020700A1",
            "US20110225644A1",
            "US20050120054A1",
            "US20050188423A1",
        ]
        data = USApplication.objects.filter(app_early_pub_number=nos)
        assert len(list(data)) == 5

    # @pytest.mark.skip('This function doesn\'nt work with the ORM Style. Consider dropping it')
    def test_mixed_get_many(self):
        pats = ["7627658", "7551922", "7359935"]
        pubs = ["US20080034424A1", "US20100020700A1", "US20110225644A1"]
        apps = ["14971450", "15332765", "13441334"]
        data = USApplication.objects.get_many(
            app_early_pub_number=pubs, patent_number=pats, appl_id=apps
        )
        assert len(data) == 9

    def test_get_search_fields(self):
        result = USApplication.objects.fields()
        assert "patent_number" in result
        assert "appl_id" in result
        assert "app_early_pub_number" in result



import re
from collections import namedtuple

from lxml import etree as ET

# XML Namespaces for ElementTree
NS = {
    "ops": "http://ops.epo.org",
    "epo": "http://www.epo.org/exchange",
    "ft": "http://www.epo.org/fulltext",
    "reg": "http://www.epo.org/register",
}
DocDB = namedtuple("DocDB", ["country", "number", "kind", "date", "doc_type"])
EpoDoc = namedtuple("EpoDoc", ["number", "kind", "date"])
whitespace_re = re.compile(" +")
country_re = re.compile(r"^[A-Z]{2}")
ep_case_re = re.compile(r"EP(?P<number>[\d]+)(?P<kind>[A-Z]\d)?")


class OPSParser:
    """Internal xml parsing class for common tasks"""

    def docdb_number(self, el, doc_type):
        raw_data = {
            "country": el.find("./epo:country", NS),
            "number": el.find("./epo:doc-number", NS),
            "kind": el.find("./epo:kind", NS),
            "date": el.find("./epo:date", NS),
        }
        for k, v in raw_data.items():
            if v is not None:
                raw_data[k] = v.text

        raw_data["doc_type"] = doc_type

        return DocDB(**raw_data)

    def epodoc_number(self, el):
        raw_data = {
            "number": el.find("./epo:doc-number", NS),
            "kind": el.find("./epo:kind", NS),
            "date": el.find("./epo:date", NS),
        }
        for k, v in raw_data.items():
            if v is not None:
                raw_data[k] = v.text

        return EpoDoc(**raw_data)

    def cpc_class(self, el):
        keys = "section, class, subclass, main-group, subgroup".split(", ")
        data = {k: el.find("./epo:" + k, NS) for k in keys}
        data = {k: v.text for (k, v) in data.items() if v is not None}
        return f"{data.get('section', '')}{data.get('class', '')}{data.get('subclass', '')} {data.get('main-group', '')}/{data.get('subgroup', '')}"

    def pct_to_docdb(self, number):
        _, country_year, number = number.split("/")
        country, year = country_year[:2], country_year[2:]
        if len(year) == 2:
            if int(year) > 50:
                year = "19" + year
            else:
                year = "20" + year

        # DocDB format changed in 2004:
        # Pre-2003 - CCyynnnnnW
        # Post-2003 - CCccyynnnnnnW
        if int(year) >= 2004:
            case_number = year + number.rjust(6, "0")
        else:
            case_number = year[2:] + number.rjust(5, "0")
        return DocDB(country, case_number, "W", None, "application")

    def citation(self, el):
        phase = el.attrib["cited-phase"]
        cited_by = el.attrib["cited-by"]
        office = el.attrib.get("office", "")
        pat_cite = el.find("./epo:patcit", NS)
        if pat_cite is not None:
            citation = dict(
                self.docdb_number(
                    pat_cite.find('./epo:document-id[@document-id-type="docdb"]', NS),
                    "publication",
                )._asdict()
            )
        else:
            citation = el.find("./epo:nplcit/epo:text", NS).text
        category = (
            el.find("./epo:category", NS).text
            if el.find("./epo:category", NS) is not None
            else ""
        )
        relevant_claims = el.find("./epo:rel-claims", NS)
        relevant_passages = [
            e.text for e in el.findall("./epo:rel-passage/epo:passage", NS)
        ]
        return {
            "phase": phase,
            "cited_by": cited_by,
            "office": office,
            "citation": citation,
            "category": category,
            "relevant_claims": relevant_claims.text
            if relevant_claims is not None
            else "",
            "relevant_passages": relevant_passages,
        }


class InpadocParser(OPSParser):
    """
    INPADOC Publication

    This is a lazy-loading object that accesses INPADOC data for the associated class.
    
    The first time any data is accessed, the result is fetched and then cached on the object.

    Data includes:
        bib_data => Basic bibliogrpahic information + Abstract
        description => Full Text Specification
        claims => Claims text
        images => Images listing
        download_images() => download a PDF of the published document
        legal => Status Information

    """

    def __init__(self, manager):
        self.manager = manager

    def legal(self, doc_db):
        tree = self.manager.xml_data(doc_db, "legal")
        legal_events = tree.findall(".//ops:legal", NS)
        output = list()
        for le in legal_events:
            row = dict()
            row["description"] = le.attrib["desc"]
            row["explanation"] = le.attrib["infl"]
            row["code"] = le.attrib["code"].strip()
            row["date"] = le.find("./ops:L007EP", NS).text
            output.append(row)
        output = list(sorted(output, key=lambda x: x["date"]))
        return output

    def bib_data(self, doc_db):
        # NOTE: For EP cases with search reports, there is citation data in bib data
        # We just currently are not parsing it
        tree = self.manager.xml_data(doc_db, "biblio")
        documents = tree.findall("./epo:exchange-documents/epo:exchange-document", NS)

        data_items = list()
        for document in documents:
            data = dict()
            data["family_id"] = document.attrib["family-id"]
            bib_data = document.find("./epo:bibliographic-data", NS)

            title = bib_data.find("./epo:invention-title[@lang='en']", NS)
            if title == None:
                title = bib_data.find("./epo:invention-title", NS)
            if title != None:
                data["title"] = title.text.strip()
            else:
                data["title"] = ""
            pub_data = bib_data.find(
                './epo:publication-reference/epo:document-id[@document-id-type="docdb"]',
                NS,
            )
            pub_data = self.manager.parser.docdb_number(pub_data, "publication")
            data["doc_db"] = pub_data
            data["publication"] = pub_data.country + pub_data.number + pub_data.kind
            data["publication_date"] = pub_data.date
            data["country"] = pub_data.country

            app_data = bib_data.find(
                './epo:application-reference/epo:document-id[@document-id-type="docdb"]',
                NS,
            )
            app_data = dict(
                self.manager.parser.docdb_number(app_data, "application")._asdict()
            )
            data["application"] = app_data["country"] + app_data["number"]
            data["filing_date"] = app_data.get("date", None)
            original_app_data = bib_data.find(
                './epo:application-reference/epo:document-id[@document-id-type="original"]',
                NS,
            )
            if original_app_data is not None:
                data["original_application_number"] = original_app_data.find(
                    "./epo:doc-number", NS
                ).text

            original_app = bib_data.find(
                './epo:application-reference/epo:document-id[@document-id-type="original"]/epo:doc-number',
                NS,
            )
            if original_app is not None:
                data["original_application_number"] = original_app.text

            intl_class = [
                whitespace_re.sub("", e.text)
                for e in bib_data.findall(
                    "./epo:classifications-ipcr/epo:classification-ipcr/epo:text", NS
                )
            ]
            data["intl_class"] = intl_class

            cpc_classes = bib_data.findall(
                "./epo:patent-classifications/epo:patent-classification", NS
            )
            data["cpc_class"] = [
                self.manager.parser.cpc_class(el) for el in cpc_classes
            ]

            priority_apps = bib_data.findall(
                './epo:priority-claims/epo:priority-claim/epo:document-id[@document-id-type="original"]/epo:doc-number',
                NS,
            )
            data["priority_claims"] = [e.text for e in priority_apps]

            parties = bib_data.find("./epo:parties", NS)
            data["applicants"] = [
                e.text
                for e in parties.findall(
                    './epo:applicants/epo:applicant[@data-format="original"]/epo:applicant-name/epo:name',
                    NS,
                )
            ]
            data["inventors"] = [
                e.text
                for e in parties.findall(
                    './epo:inventors/epo:inventor[@data-format="original"]/epo:inventor-name/epo:name',
                    NS,
                )
            ]
            abstract = document.find('./epo:abstract[@lang="en"]', NS)
            data["abstract"] = (
                "".join(t for t in abstract.itertext()).strip()
                if abstract is not None
                else ""
            )

            refs_cited = bib_data.findall("./epo:references-cited/epo:citation", NS)
            data["references_cited"] = [
                self.manager.parser.citation(c) for c in refs_cited
            ]
            data_items.append(data)
        return data_items

    def description(self, doc_db):
        tree = self.manager.xml_data(doc_db, "description")
        description = tree.find(
            "./ft:fulltext-documents/ft:fulltext-document/ft:description", NS
        )
        text = "\n".join(description.itertext()).strip()
        return text

    def claims(self, doc_db):
        tree = self.manager.xml_data(doc_db, "claims")
        claims = [
            "".join(e.itertext()).strip()
            for e in tree.findall(
                './ft:fulltext-documents/ft:fulltext-document/ft:claims[@lang="EN"]/ft:claim/ft:claim-text',
                NS,
            )
        ]
        return claims

    def images(self, doc_db):
        if doc_db.doc_type == "application":
            return dict()
        tree = self.manager.xml_data(doc_db, "images")
        images = tree.find(
            './ops:document-inquiry/ops:inquiry-result/ops:document-instance[@desc="FullDocument"]',
            NS,
        )
        data = {
            "url": "http://ops.epo.org/rest-services/" + images.attrib["link"] + ".pdf",
            "num_pages": int(images.attrib["number-of-pages"]),
            "sections": {
                e.attrib["name"]: int(e.attrib["start-page"])
                for e in images.findall("./ops:document-section", NS)
            },
        }
        return data

    def family(self, doc_db):
        tree = self.manager.xml_data(doc_db, "family")
        doc_db_list = list()
        for el in tree.findall("./ops:patent-family/ops:family-member", NS):
            pub_ref = el.find(
                './epo:publication-reference/epo:document-id[@document-id-type="docdb"]',
                NS,
            )
            if pub_ref is not None:
                doc_db_list.append(
                    self.manager.parser.docdb_number(pub_ref, "publication")
                )
            else:
                app_ref = el.find(
                    './epo:application-reference/epo:document-id[@document-id-type="docdb"]',
                    NS,
                )
                doc_db_list.append(
                    self.manager.parser.docdb_number(app_ref, "application")
                )
        return doc_db_list



import json
import os
import re
import shutil
import time
import xml.etree.ElementTree as ET

import requests
from patent_client import CACHE_BASE
from patent_client import SETTINGS

from .util import Manager
from .util import Model
from .util import one_to_many
from .util import one_to_one

CLIENT_SETTINGS = SETTINGS["ItcEdis"]
if os.environ.get("EDIS_USER", False):
    USERNAME = os.environ["EDIS_USER"]
    PASSWORD = os.environ["EDIS_PASS"]
else:
    USERNAME = CLIENT_SETTINGS["Username"]
    PASSWORD = CLIENT_SETTINGS["Password"]

BASE_URL = "https://edis.usitc.gov/data"
CACHE_DIR = CACHE_BASE / "itc_edis"
CACHE_DIR.mkdir(exist_ok=True)

session = requests.Session()

# API Guide at: https://www.usitc.gov/docket_services/documents/EDIS3WebServiceGuide.pdf


class ITCInvestigationManager(Manager):
    max_retries = 3
    auth_time = 10 * 60  # Re-authenticate every # seconds
    last_auth = 0
    base_url = BASE_URL + "/investigation/"

    def authenticate():
        if (
            time.time() - ITCInvestigationManager.last_auth
            > ITCInvestigationManager.auth_time
        ):
            path = "/secretKey/" + USERNAME
            response = session.get(BASE_URL + path, params={"password": PASSWORD})
            tree = ET.fromstring(response.text)
            key = tree.find("secretKey").text
            session.auth = (USERNAME, key)

    def filter(self):
        raise NotImplementedError("EDIS Api does not have a search function!")

    def get(self, investigation_number):
        fname = CACHE_DIR / (investigation_number + ".json")
        if fname.exists():
            return ITCInvestigation(json.load(open(fname)))
        else:
            ITCInvestigationManager.authenticate()
            url = self.base_url + investigation_number
            response = session.get(url)
            tree = ET.fromstring(response.text)
            tree = tree[0][0]
            data = {
                "phase": tree.find("investigationPhase").text,
                "number": tree.find("investigationNumber").text,
                "status": tree.find("investigationStatus").text,
                "title": tree.find("investigationTitle").text,
                "type": tree.find("investigationType").text,
                "doc_list_url": tree.find("documentListUri").text,
                "docket_number": tree.find("docketNumber").text,
            }
            with open(fname, "w") as f:
                json.dump(data, f, indent=2)
            return ITCInvestigation(data)


class ITCInvestigation(Model):
    objects = ITCInvestigationManager()
    documents = one_to_many("patent_client.ITCDocument", investigation_number="number")

    def __repr__(self):
        return f"<ITCInvestigation(number={self.number})>"


class ITCDocumentsManager(Manager):
    primary_key = "investigation_number"
    base_url = BASE_URL + "/document"
    allowed_filters = {
        "investigation_number": "investigationNumber",
        "phase": "investigationPhase",
        "type": "documentType",
        "firm": "firmOrg",
        "security": "securityLevel",
    }

    def get(self, document_id):
        fname = CACHE_DIR / f"document-{document_id}.json"
        if not fname.exists():
            ITCInvestigationManager.authenticate()
            response = session.get(f"{self.base_url}/{document_id}")
            tree = ET.fromstring(response.text)
            doc_el = tree.find(".//document")
            data = self.parse_doc(doc_el)
            with open(fname, "w") as f:
                json.dump(data, f, indent=2)
        else:
            data = json.load(open(fname))
        return ITCDocument(data)

    def parse_doc(self, element):
        attribute_dict = dict(
            type="documentType",
            title="documentTitle",
            investigation_number="investigationNumber",
            security="securityLevel",
            filing_org="firmOrganization",
            filed_by="filedBy",
            filed_on_behalf_of="onBehalfOf",
            action_jacket_control_number="actionJacketControlNumber",
            memorandum_control_number="memorandumControlNumber",
            attachment_url="attachmentListUri",
            date="documentDate",
            last_modified="modifiedDate",
            id="id",
        )
        data = dict()
        for key, value in attribute_dict.items():
            data[key] = element.find(value).text
        return data

    def get_item(self, key):
        query = {self.allowed_filters[k]: v for (k, v) in self.filter_params.items()}
        page = int(key / 100) + 1
        location = key % 100
        query["pagenumber"] = page
        q_string = re.sub(r'[\{\}":, ]+', "-", json.dumps(query, sort_keys=True)[1:-1])
        fname = CACHE_DIR / f"document-page-{q_string}.json"
        if fname.exists():
            data = json.load(open(fname))
        else:
            ITCInvestigationManager.authenticate()
            response = session.get(self.base_url, params=query)
            tree = ET.fromstring(response.text)[0]
            data = list()
            for element in tree.findall("document"):
                data.append(self.parse_doc(element))
            with open(fname, "w") as f:
                json.dump(data, f, indent=2)
        return ITCDocument(data[location])


class ITCDocument(Model):
    objects = ITCDocumentsManager()
    investigation = one_to_one(
        "patent_client.ITCInvestigation", investigation_number="investigation_number"
    )
    attachments = one_to_many("patent_client.ITCAttachment", document_id="id")

    def __repr__(self):
        return f"<ITCDocument(title={self.title})>"


class ITCAttachmentManager(Manager):
    primary_key = "investigation_number"
    base_url = BASE_URL + "/attachment/"
    allowed_filters = ["document_id"]

    def get_item(self, key):
        doc_id = self.filter_params["document_id"]
        fname = CACHE_DIR / f"attachments-{doc_id}.json"
        if fname.exists() and False:
            data = json.load(open(fname))
        else:
            ITCInvestigationManager.authenticate()
            response = session.get(self.base_url + doc_id)
            tree = ET.fromstring(response.text)
            attribute_dict = dict(
                id="id",
                document_id="documentId",
                title="title",
                size="fileSize",
                file_name="originalFileName",
                pages="pageCount",
                created_date="createDate",
                last_modified_date="lastModifiedDate",
                download_url="downloadUri",
            )
            data = list()
            for element in tree.findall(".//attachment"):
                row = dict()
                for k, value in attribute_dict.items():
                    row[k] = element.find(value).text
                data.append(row)
            with open(fname, "w") as f:
                json.dump(data, f, indent=2)
        return ITCAttachment(data[key])


class ITCAttachment(Model):
    objects = ITCAttachmentManager()
    document = one_to_one("patent_client.ITCDocument", document_id="document_id")

    def __repr__(self):
        return f"<ITCAttachment(title={self.title})>"

    def download(self, path="."):
        *_, ext = self.file_name.split(".")
        filename = f"{self.document.title.strip()} - {self.title}.{ext}"
        cdir = os.path.join(CACHE_DIR, self.document.investigation.number)
        os.makedirs(cdir, exist_ok=True)
        cname = os.path.join(cdir, filename)
        oname = os.path.join(path, filename)
        if not os.path.exists(cname):
            response = session.get(self.download_url, stream=True)
            with open(cname, "wb") as f:
                for chunk in response.iter_content(1024):
                    f.write(chunk)
        shutil.copy(cname, oname)
